/*
 This example code is in the Public Domain (or CC0 licensed, at your option.)

 Unless required by applicable law or agreed to in writing, this
 software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 CONDITIONS OF ANY KIND, either express or implied.
 */

/****************************************************************************
 *
 * This file is for ble spp client demo.
 *
 ****************************************************************************/

#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <stdio.h>
#include "driver/uart.h"

#include "esp_bt.h"
#include "nvs_flash.h"
#include "esp_bt_device.h"
#include "esp_gap_ble_api.h"
#include "esp_gattc_api.h"
#include "esp_gatt_defs.h"
#include "esp_bt_main.h"
#include "esp_system.h"
#include "esp_gatt_common_api.h"
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
//#include "bta_gattc_int.h"

#define GATTC_TAG                   "GATTC_SPP_DEMO"
#define PROFILE_NUM                 1
#define PROFILE_APP_ID              0
#define ESP_GATT_SPP_SERVICE_UUID   0xABF0
#define SCAN_ALL_THE_TIME           0

esp_bd_addr_t remote_bda;

uint8_t write_val[] = "Hello";

///Declare static functions
static void esp_gap_cb (esp_gap_ble_cb_event_t event,
        esp_ble_gap_cb_param_t *param);
static void gattc_profile_event_handler (esp_gattc_cb_event_t event,
        esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param);
static esp_ble_scan_params_t ble_scan_params = {.scan_type =
        BLE_SCAN_TYPE_ACTIVE, .own_addr_type = BLE_ADDR_TYPE_RPA_PUBLIC,
        .scan_filter_policy = BLE_SCAN_FILTER_ALLOW_ALL, .scan_interval = 0x50,
        .scan_window = 0x30, .scan_duplicate = BLE_SCAN_DUPLICATE_DISABLE};

static const char device_name[] = "ESP_BLE_SECURITY";
static uint16_t client_conn_id = 0;
static uint16_t client_mtu_size = 23;
static uint16_t cmd = 0;
static uint16_t srv_start_handle = 0;
static uint16_t srv_end_handle = 0;
static uint16_t client_gattc_if = 0xff;
static uint16_t service_count = 0;
static uint16_t char_count = 0;
static uint16_t count = 0;
static esp_gattc_db_elem_t *db = NULL;
static esp_ble_gap_cb_param_t scan_rst;

static esp_bt_uuid_t service_uuid = {.len = ESP_UUID_LEN_16, .uuid = {.uuid16 =
ESP_GATT_SPP_SERVICE_UUID, }, };

static void free_gattc_srv_db (void)
{
    client_gattc_if = 0xff;
    client_conn_id = 0;
    client_mtu_size = 23;
    cmd = 0;
    srv_start_handle = 0;
    srv_end_handle = 0;
    if (db) {
        free(db);
        db = NULL;
    }
}

static void show_bonded_devices (void)
{
    int dev_num = esp_ble_get_bond_device_num();

    if (dev_num) {
        esp_ble_bond_dev_t *dev_list = (esp_ble_bond_dev_t *)malloc(
                sizeof(esp_ble_bond_dev_t) * dev_num);
        esp_ble_get_bond_device_list(&dev_num, dev_list);
        ESP_LOGI(GATTC_TAG, "Bonded devices number : %d\n", dev_num);

        ESP_LOGI(GATTC_TAG, "Bonded devices list : %d\n", dev_num);
        for (int i = 0; i < dev_num; i++) {
            esp_log_buffer_hex(GATTC_TAG, (void *)dev_list[i].bd_addr, sizeof(esp_bd_addr_t));
        }

        free(dev_list);
    }
}

void ble_key_info (esp_ble_key_t *ble_key)
{

    esp_bd_addr_t bd_addr;
    uint8_t index = 0;
    memcpy(bd_addr, ble_key->bd_addr, sizeof(esp_bd_addr_t));
    ESP_LOGI(GATTC_TAG, "remote BD_ADDR: %08x%04x",
            (bd_addr[0] << 24) + (bd_addr[1] << 16) + (bd_addr[2] << 8)
                    + bd_addr[3], (bd_addr[4] << 8) + bd_addr[5]);
    switch (ble_key->key_type) {
        case ESP_LE_KEY_NONE:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_NONE");
            break;
        case ESP_LE_KEY_PENC:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_PENC");
            ESP_LOGI(GATTC_TAG, "Peer LTK Value");
            for (; index < 16; index++) {
                ESP_LOGI(GATTC_TAG, "%x",
                        ble_key->p_key_value.penc_key.ltk[index]);
            }
            ESP_LOGI(GATTC_TAG, "Peer Random Value");
            for (index = 0; index < 16; index++) {
                ESP_LOGI(GATTC_TAG, "%x",
                        ble_key->p_key_value.penc_key.rand[index]);
            }
            ESP_LOGI(GATTC_TAG, "Peer EDIV:%x",
                    ble_key->p_key_value.penc_key.ediv);
            ESP_LOGI(GATTC_TAG, "Security Level:%d",
                    ble_key->p_key_value.penc_key.sec_level);
            ESP_LOGI(GATTC_TAG, "Key Size:%d",
                    ble_key->p_key_value.penc_key.key_size);
            break;
        case ESP_LE_KEY_PID:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_PID");
            ESP_LOGI(GATTC_TAG, "Peer ID Key");
            for (index = 0; index < 16; index++) {
                ESP_LOGI(GATTC_TAG, "%x",
                        ble_key->p_key_value.pid_key.irk[index]);
            }
            ESP_LOGI(GATTC_TAG, "Address Type:%d",
                    ble_key->p_key_value.pid_key.addr_type);
            ESP_LOGI(GATTC_TAG, "Static Address");
//            for (index = 0; index < 6; index++) {
//                ESP_LOGI(GATTC_TAG, "%x",
//                        ble_key->p_key_value.pid_key.static_addr[index]);
//            }

            esp_log_buffer_hex(GATTC_TAG, (void *)ble_key->p_key_value.pid_key.static_addr, sizeof(esp_bd_addr_t));
            break;
        case ESP_LE_KEY_PCSRK:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_PCSRK");
            for (index = 0; index < 16; index++) {
                ESP_LOGI(GATTC_TAG, "%x",
                        ble_key->p_key_value.pcsrk_key.csrk[index]);
            }
            break;
        case ESP_LE_KEY_PLK:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_PLK");
            break;
        case ESP_LE_KEY_LLK:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_LLK");
            break;
        case ESP_LE_KEY_LENC:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_LENC");
            ESP_LOGI(GATTC_TAG, "Local LTK Value");
            for (; index < 16; index++) {
                ESP_LOGI(GATTC_TAG, "%x",
                        ble_key->p_key_value.lenc_key.ltk[index]);
            }
            ESP_LOGI(GATTC_TAG, "Local EDIV:%x",
                    ble_key->p_key_value.lenc_key.div);
            ESP_LOGI(GATTC_TAG, "Security Level:%d",
                    ble_key->p_key_value.lenc_key.sec_level);
            ESP_LOGI(GATTC_TAG, "Key Size:%d",
                    ble_key->p_key_value.lenc_key.key_size);
            break;
        case ESP_LE_KEY_LID:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_LID");
            break;
        case ESP_LE_KEY_LCSRK:
            ESP_LOGI(GATTC_TAG, "ESP_LE_KEY_LCSRK");
            break;
        default:
            ESP_LOGI(GATTC_TAG, "INVALID BLE KEY TYPE");
            break;
    }
}

static char *esp_auth_req_to_str (esp_ble_auth_req_t auth_req)
{
    char *auth_str = NULL;
    switch (auth_req) {
        case ESP_LE_AUTH_NO_BOND:
            auth_str = "ESP_LE_AUTH_NO_BOND";
            break;
        case ESP_LE_AUTH_BOND:
            auth_str = "ESP_LE_AUTH_BOND";
            break;
        case ESP_LE_AUTH_REQ_MITM:
            auth_str = "ESP_LE_AUTH_REQ_MITM";
            break;
        case ESP_LE_AUTH_REQ_SC_ONLY:
            auth_str = "ESP_LE_AUTH_REQ_SC_ONLY";
            break;
        case ESP_LE_AUTH_REQ_SC_BOND:
            auth_str = "ESP_LE_AUTH_REQ_SC_BOND";
            break;
        case ESP_LE_AUTH_REQ_SC_MITM:
            auth_str = "ESP_LE_AUTH_REQ_SC_MITM";
            break;
        case ESP_LE_AUTH_REQ_SC_MITM_BOND:
            auth_str = "ESP_LE_AUTH_REQ_SC_MITM_BOND";
            break;
        default:
            auth_str = "INVALID BLE AUTH REQ";
            break;
    }

    return auth_str;
}

static void remove_all_bonded_devices (void)
{
    int dev_num = esp_ble_get_bond_device_num();

    if (dev_num) {
    ESP_LOGI(GATTC_TAG, "Removing Bonded Device");
    esp_ble_bond_dev_t *dev_list = (esp_ble_bond_dev_t *)malloc(
            sizeof(esp_ble_bond_dev_t) * dev_num);
    esp_ble_get_bond_device_list(&dev_num, dev_list);
    for (int i = 0; i < dev_num; i++) {
        esp_ble_remove_bond_device(dev_list[i].bd_addr);
    }

    free(dev_list);
}
}

static void esp_gap_cb (esp_gap_ble_cb_event_t event,
        esp_ble_gap_cb_param_t *param)
{
    uint8_t *adv_name = NULL;
    uint8_t adv_name_len = 0;
    esp_err_t err;

    ESP_LOGI(GATTC_TAG, "GAP_EVT, event %d\n", event);

    switch (event) {
        case ESP_GAP_BLE_SCAN_PARAM_SET_COMPLETE_EVT: {
            if ((err = param->scan_param_cmpl.status)
                    != ESP_BT_STATUS_SUCCESS) {
                ESP_LOGE(GATTC_TAG, "Scan param set failed: %s",
                        esp_err_to_name(err));
                break;
            }
            //the unit of the duration is second
            uint32_t duration = 0xFFFF;
            ESP_LOGI(GATTC_TAG, "Enable Ble Scan:during time 0x%04X minutes.",
                    duration);
            esp_ble_gap_start_scanning(duration);
            break;
        }
        case ESP_GAP_BLE_SCAN_START_COMPLETE_EVT:
            //scan start complete event to indicate scan start successfully or failed
            if ((err = param->scan_start_cmpl.status)
                    != ESP_BT_STATUS_SUCCESS) {
                ESP_LOGE(GATTC_TAG, "Scan start failed: %s",
                        esp_err_to_name(err));
                break;
            }
            ESP_LOGI(GATTC_TAG, "Scan start successed");
            break;
        case ESP_GAP_BLE_SCAN_STOP_COMPLETE_EVT:
            if ((err = param->scan_stop_cmpl.status) != ESP_BT_STATUS_SUCCESS) {
                ESP_LOGE(GATTC_TAG, "Scan stop failed: %s",
                        esp_err_to_name(err));
                break;
            }
            ESP_LOGI(GATTC_TAG, "Scan stop successed");
            esp_ble_gattc_open(client_gattc_if, scan_rst.scan_rst.bda,
                    scan_rst.scan_rst.ble_addr_type,
                    true);
            break;
        case ESP_GAP_BLE_SCAN_RESULT_EVT: {
            esp_ble_gap_cb_param_t *scan_result =
                    (esp_ble_gap_cb_param_t *)param;
            switch (scan_result->scan_rst.search_evt) {
                case ESP_GAP_SEARCH_INQ_RES_EVT:
                    esp_log_buffer_hex(GATTC_TAG, scan_result->scan_rst.bda, 6);
                    ESP_LOGI(GATTC_TAG,
                            "Searched Adv Data Len %d, Scan Response Len %d",
                            scan_result->scan_rst.adv_data_len,
                            scan_result->scan_rst.scan_rsp_len);
                    adv_name = esp_ble_resolve_adv_data(
                            scan_result->scan_rst.ble_adv,
                            ESP_BLE_AD_TYPE_NAME_CMPL, &adv_name_len);
                    ESP_LOGI(GATTC_TAG, "Searched Device Name Len %d",
                            adv_name_len);
                    esp_log_buffer_char(GATTC_TAG, adv_name, adv_name_len);
                    ESP_LOGI(GATTC_TAG, "\n");
                    if (adv_name != NULL) {
                        if (strncmp((char *)adv_name, device_name, adv_name_len)
                                == 0) {
                            memcpy(&(scan_rst), scan_result,
                                    sizeof(esp_ble_gap_cb_param_t));
                            esp_ble_gap_stop_scanning();
                        }
                    }
                    break;
                case ESP_GAP_SEARCH_INQ_CMPL_EVT:
                    break;
                default:
                    break;
            }
            break;
        }
        case ESP_GAP_BLE_ADV_STOP_COMPLETE_EVT:
            if ((err = param->adv_stop_cmpl.status) != ESP_BT_STATUS_SUCCESS) {
                ESP_LOGE(GATTC_TAG, "Adv stop failed: %s",
                        esp_err_to_name(err));
            } else {
                ESP_LOGI(GATTC_TAG, "Stop adv successfully");
            }
            break;
        case ESP_GAP_BLE_SEC_REQ_EVT:
            /* send the positive(true) security response to the peer device to accept the security request.
             If not accept the security request, should sent the security response with negative(false) accept value*/
            esp_ble_gap_security_rsp(param->ble_security.ble_req.bd_addr, true);
            break;

        case ESP_GAP_BLE_AUTH_CMPL_EVT: {
            esp_bd_addr_t bd_addr;
            memcpy(bd_addr, param->ble_security.auth_cmpl.bd_addr,
                    sizeof(esp_bd_addr_t));
            ESP_LOGI(GATTC_TAG, "remote BD_ADDR: %08x%04x",
                    (bd_addr[0] << 24) + (bd_addr[1] << 16) + (bd_addr[2] << 8)
                            + bd_addr[3], (bd_addr[4] << 8) + bd_addr[5]);
            ESP_LOGI(GATTC_TAG, "address type = %d",
                    param->ble_security.auth_cmpl.addr_type);
            ESP_LOGI(GATTC_TAG, "pair status = %s",
                    param->ble_security.auth_cmpl.success ? "success" : "fail");
            if (!param->ble_security.auth_cmpl.success) {
                ESP_LOGI(GATTC_TAG, "fail reason = 0x%x",
                        param->ble_security.auth_cmpl.fail_reason);
            } else {
                ESP_LOGI(GATTC_TAG, "auth mode = %s",
                        esp_auth_req_to_str(
                                param->ble_security.auth_cmpl.auth_mode));
            }
            show_bonded_devices();
            break;
        }
        case ESP_GAP_BLE_SET_LOCAL_PRIVACY_COMPLETE_EVT:
            esp_ble_gap_set_scan_params(&ble_scan_params);
            break;

        case ESP_GAP_BLE_KEY_EVT: {
            //shows the ble key info share with peer device to the user.

            ble_key_info(&(param->ble_security.ble_key));
            //        ESP_LOGI(GATTC_TAG, "key type = %s", esp_key_type_to_str(param->ble_security.ble_key.key_type));
        }
            break;
        default:
            break;
    }
}

static void gattc_profile_event_handler (esp_gattc_cb_event_t event,
        esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param)
{
    esp_ble_gattc_cb_param_t *p_data = (esp_ble_gattc_cb_param_t *)param;
    uint16_t att_count = 0;
    uint16_t char_count = 0;
    uint16_t descr_count;
    uint32_t ret;

    switch (event) {
        case ESP_GATTC_REG_EVT:
            ESP_LOGI(GATTC_TAG, "REG EVT, set scan params");
            esp_ble_gap_config_local_privacy(true);
            client_gattc_if = gattc_if;
            break;
        case ESP_GATTC_CONNECT_EVT:
            client_conn_id = p_data->connect.conn_id;
            memcpy(remote_bda, p_data->connect.remote_bda,
                    sizeof(esp_bd_addr_t));
            ESP_LOGI(GATTC_TAG, "REMOTE BDA:");
            esp_log_buffer_hex(GATTC_TAG, remote_bda, sizeof(esp_bd_addr_t));

            //esp_ble_gattc_search_service(client_gattc_if, client_conn_id,
            //      &service_uuid);
#if 0
            //esp_ble_gattc_get_attr_count(spp_gattc_if, spp_conn_id,
            //        ESP_GATT_DB_PRIMARY_SERVICE, 1, 65535, 0, &service_count);
            //ESP_LOGI(GATTC_TAG, "Service Value:%d\n", service_count);
            //ret = esp_ble_gattc_get_attr_count(spp_gattc_if, spp_conn_id, ESP_GATT_DB_PRIMARY_SERVICE, 1,
            //                              65535, 0, &count);
            //        ESP_LOGI(GATTC_TAG, "Return Value:%d\n", ret);
            //      ESP_LOGI(GATTC_TAG, "Service Count:%d\n", count);
#endif
            break;
        case ESP_GATTC_DISCONNECT_EVT:
            ESP_LOGI(GATTC_TAG, "disconnect");
            free_gattc_srv_db();
            esp_ble_gap_start_scanning(SCAN_ALL_THE_TIME);
            break;
        case ESP_GATTC_SEARCH_RES_EVT:
            ESP_LOGI(GATTC_TAG,
                    "ESP_GATTC_SEARCH_RES_EVT: start_handle = %d, end_handle = %d, UUID:0x%04x",
                    p_data->search_res.start_handle,
                    p_data->search_res.end_handle,
                    p_data->search_res.srvc_id.uuid.uuid.uuid16);
            srv_start_handle = p_data->search_res.start_handle;
            srv_end_handle = p_data->search_res.end_handle;
#if 0
            esp_ble_gattc_get_attr_count(spp_gattc_if, spp_conn_id,
                    ESP_GATT_DB_CHARACTERISTIC, srv_start_handle,
                    srv_end_handle, 0, &count);
            ESP_LOGI(GATTC_TAG, "Char Value:%d\n", count);
            esp_ble_gattc_get_attr_count(spp_gattc_if, spp_conn_id,
                    ESP_GATT_DB_ALL, srv_start_handle, srv_end_handle,
                    0, &count);
            ESP_LOGI(GATTC_TAG, "Full Database Value:%d\n", count);
            esp_ble_gattc_get_attr_count(spp_gattc_if, spp_conn_id,
                    ESP_GATT_DB_PRIMARY_SERVICE, 1, 65535, 0, &count);
            ESP_LOGI(GATTC_TAG, "Service Value:%d\n", count);
#endif
            break;
        case ESP_GATTC_SEARCH_CMPL_EVT:
            ESP_LOGI(GATTC_TAG, "SEARCH_CMPL: conn_id = %x, status %d",
                    client_conn_id, p_data->search_cmpl.status);

            esp_ble_gattc_get_attr_count(client_gattc_if, client_conn_id,
                    ESP_GATT_DB_ALL, srv_start_handle, srv_end_handle, 0,
                    &att_count);
            ESP_LOGI(GATTC_TAG, "Att count:%d\n", att_count);

            db = (esp_gattc_db_elem_t *)malloc(
                    (att_count + 2) * sizeof(esp_gattc_db_elem_t));
            if (db == NULL) {
                ESP_LOGE(GATTC_TAG, "%s:malloc db falied\n", __func__);
                break;
            }
            count = att_count;
            ret = esp_ble_gattc_get_db(client_gattc_if, client_conn_id,
                    srv_start_handle, srv_end_handle, db, &count);
            ESP_LOGE(GATTC_TAG, "Return Error:%d", ret);
            if (ret != ESP_GATT_OK) {
                ESP_LOGE(GATTC_TAG, "%s:get db falied\n", __func__);
                break;
            }
            ESP_LOGI(GATTC_TAG,
                    "Att count after get the database from Cache Memory: %d\n",
                    count);
            if (count != att_count) {
                ESP_LOGE(GATTC_TAG, "Att count does not match");
                break;
            }
            for (int i = 0; i < att_count; i++) {
                switch ((db + i)->type) {
                    case ESP_GATT_DB_PRIMARY_SERVICE:
                        ESP_LOGI(GATTC_TAG,
                                "attr_type = PRIMARY_SERVICE,attribute_handle=%d,start_handle=%d,end_handle=%d,properties=0x%x,uuid=0x%04x\n",
                                (db + i)->attribute_handle,
                                (db + i)->start_handle, (db + i)->end_handle,
                                (db + i)->properties,
                                (db + i)->uuid.uuid.uuid16);
                        break;
                    case ESP_GATT_DB_SECONDARY_SERVICE:
                        ESP_LOGI(GATTC_TAG,
                                "attr_type = SECONDARY_SERVICE,attribute_handle=%d,start_handle=%d,end_handle=%d,properties=0x%x,uuid=0x%04x\n",
                                (db + i)->attribute_handle,
                                (db + i)->start_handle, (db + i)->end_handle,
                                (db + i)->properties,
                                (db + i)->uuid.uuid.uuid16);
                        break;
                    case ESP_GATT_DB_CHARACTERISTIC:
                        ESP_LOGI(GATTC_TAG,
                                "attr_type = CHARACTERISTIC,attribute_handle=%d,start_handle=%d,end_handle=%d,properties=0x%x,uuid=0x%04x\n",
                                (db + i)->attribute_handle,
                                (db + i)->start_handle, (db + i)->end_handle,
                                (db + i)->properties,
                                (db + i)->uuid.uuid.uuid16);
                        if ((db + i)->properties == 0x04) {
                            esp_ble_gattc_write_char(client_gattc_if,
                                    client_conn_id, (db + i)->attribute_handle,
                                    strlen((char *)write_val), write_val,
                                    ESP_GATT_WRITE_TYPE_NO_RSP,
                                    ESP_GATT_AUTH_REQ_NONE);
                        }
                        break;
                    case ESP_GATT_DB_DESCRIPTOR:
                        ESP_LOGI(GATTC_TAG,
                                "attr_type = DESCRIPTOR,attribute_handle=%d,start_handle=%d,end_handle=%d,properties=0x%x,uuid=0x%04x\n",
                                (db + i)->attribute_handle,
                                (db + i)->start_handle, (db + i)->end_handle,
                                (db + i)->properties,
                                (db + i)->uuid.uuid.uuid16);
                        break;
                    case ESP_GATT_DB_INCLUDED_SERVICE:
                        ESP_LOGI(GATTC_TAG,
                                "attr_type = INCLUDED_SERVICE,attribute_handle=%d,start_handle=%d,end_handle=%d,properties=0x%x,uuid=0x%04x\n",
                                (db + i)->attribute_handle,
                                (db + i)->start_handle, (db + i)->end_handle,
                                (db + i)->properties,
                                (db + i)->uuid.uuid.uuid16);
                        break;
                    case ESP_GATT_DB_ALL:
                        ESP_LOGI(GATTC_TAG,
                                "attr_type = ESP_GATT_DB_ALL,attribute_handle=%d,start_handle=%d,end_handle=%d,properties=0x%x,uuid=0x%04x\n",
                                (db + i)->attribute_handle,
                                (db + i)->start_handle, (db + i)->end_handle,
                                (db + i)->properties,
                                (db + i)->uuid.uuid.uuid16);
                        break;
                    default:
                        break;
                }
            }
            break;
        case ESP_GATTC_REG_FOR_NOTIFY_EVT:
            break;
        case ESP_GATTC_NOTIFY_EVT:
            ESP_LOGI(GATTC_TAG, "ESP_GATTC_NOTIFY_EVT\n");
            break;
        case ESP_GATTC_READ_CHAR_EVT:
            ESP_LOGI(GATTC_TAG, "ESP_GATTC_READ_CHAR_EVT\n");
            break;
        case ESP_GATTC_WRITE_CHAR_EVT:
            break;
        case ESP_GATTC_PREP_WRITE_EVT:
            break;
        case ESP_GATTC_EXEC_EVT:
            break;
        case ESP_GATTC_WRITE_DESCR_EVT:
            break;
        case ESP_GATTC_CFG_MTU_EVT:
            //esp_ble_gattc_get_all_char(spp_gattc_if, spp_conn_id, srv_start_handle, srv_end_handle, NULL, &char_count, 0);
            //ESP_LOGI(GATTC_TAG, "Char_count: %d\n", char_count);

            //esp_ble_gattc_get_all_descr(spp_gattc_if, spp_conn_id, srv_start_handle, srv_end_handle, NULL, &descr_count, 0);
            //ESP_LOGI(GATTC_TAG, "Descr count: %d\n", descr_count);

            //bta_gattc_get_db_size_handle(spp_gattc_if, srv_start_handle, srv_end_handle, &count);
            break;
        case ESP_GATTC_SRVC_CHG_EVT:

            break;
        default:
            break;
    }
}

void ble_client_appRegister (void)
{
    esp_err_t status;
    char err_msg[20];

//register the scan callback function to the gap module
    if ((status = esp_ble_gap_register_callback(esp_gap_cb)) != ESP_OK) {
        ESP_LOGE(GATTC_TAG, "gap register error: %s",
                esp_err_to_name_r(status, err_msg, sizeof(err_msg)));
        return;
    }
//register the callback function to the gattc module
    if ((status = esp_ble_gattc_register_callback(gattc_profile_event_handler))
            != ESP_OK) {
        ESP_LOGE(GATTC_TAG, "gattc register error: %s",
                esp_err_to_name_r(status, err_msg, sizeof(err_msg)));
        return;
    }
    esp_ble_gattc_app_register(PROFILE_APP_ID);
}

void app_main ( )
{
    esp_err_t ret;

    ESP_ERROR_CHECK(esp_bt_controller_mem_release(ESP_BT_MODE_CLASSIC_BT));

    esp_bt_controller_config_t bt_cfg = BT_CONTROLLER_INIT_CONFIG_DEFAULT()
    ;

    nvs_flash_init();
    ret = esp_bt_controller_init(&bt_cfg);
    if (ret) {
        ESP_LOGE(GATTC_TAG, "%s enable controller failed: %s\n", __func__,
                esp_err_to_name(ret));
        return;
    }

    ret = esp_bt_controller_enable(ESP_BT_MODE_BLE);
    if (ret) {
        ESP_LOGE(GATTC_TAG, "%s enable controller failed: %s\n", __func__,
                esp_err_to_name(ret));
        return;
    }

    ESP_LOGI(GATTC_TAG, "%s init bluetooth\n", __func__);
    ret = esp_bluedroid_init();
    if (ret) {
        ESP_LOGE(GATTC_TAG, "%s init bluetooth failed: %s\n", __func__,
                esp_err_to_name(ret));
        return;
    }
    ret = esp_bluedroid_enable();
    if (ret) {
        ESP_LOGE(GATTC_TAG, "%s enable bluetooth failed: %s\n", __func__,
                esp_err_to_name(ret));
        return;
    }

    vTaskDelay(30000 / portTICK_PERIOD_MS);
    remove_all_bonded_devices();

    ble_client_appRegister();

    /* set the security iocap & auth_req & key size & init key response key parameters to the stack*/
    esp_ble_auth_req_t auth_req = ESP_LE_AUTH_REQ_SC_BOND; //bonding with peer device after authentication
    esp_ble_io_cap_t iocap = ESP_IO_CAP_NONE; //set the IO capability to No output No input
    uint8_t key_size = 16;      //the key size should be 7~16 bytes
    uint8_t init_key = ESP_BLE_ID_KEY_MASK;
    uint8_t rsp_key = ESP_BLE_ID_KEY_MASK;
    esp_ble_gap_set_security_param(ESP_BLE_SM_AUTHEN_REQ_MODE, &auth_req,
            sizeof(uint8_t));
    esp_ble_gap_set_security_param(ESP_BLE_SM_IOCAP_MODE, &iocap,
            sizeof(uint8_t));
    esp_ble_gap_set_security_param(ESP_BLE_SM_MAX_KEY_SIZE, &key_size,
            sizeof(uint8_t));
    /* If your BLE device act as a Slave, the init_key means you hope which types of key of the master should distribut to you,
     and the response key means which key you can distribut to the Master;
     If your BLE device act as a master, the response key means you hope which types of key of the slave should distribut to you,
     and the init key means which key you can distribut to the slave. */
    esp_ble_gap_set_security_param(ESP_BLE_SM_SET_INIT_KEY, &init_key,
            sizeof(uint8_t));
    esp_ble_gap_set_security_param(ESP_BLE_SM_SET_RSP_KEY, &rsp_key,
            sizeof(uint8_t));
}
